#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import os
import signal
import subprocess
import psutil

import logging
logger = logging.getLogger(__name__)

from django.db import models
from django.db import transaction
from django.conf import settings
from django.core.urlresolvers import reverse

from ...common.texts import Messages


#----------------------------------------------------------


def _cleanup_zombies():
    '''Cleans-up eventual zombie subprocesses launched by the worker'''

    for child in psutil.Process().children(recursive=True):
        try:
            if child.status() == psutil.STATUS_ZOMBIE:
                child.wait()
        except psutil.NoSuchProcess:
            # process is gone meanwhile, which is ok
            pass


#----------------------------------------------------------


class WorkerManager(models.Manager):

    def get_by_natural_key(self, name):
        return self.get(name=name)


#----------------------------------------------------------


class Worker(models.Model):

    name = models.CharField(
        max_length=100,
        help_text=Messages['name'],
        unique=True,
    )

    active = models.BooleanField(
        help_text=u'If this worker is usable presently',
        default=False,
    )

    update = models.BooleanField(
        help_text=u'If this worker state must be updated at the next cycle',
        default=False,
    )

    updated = models.DateTimeField(null=True, auto_now=True)

    cores = models.PositiveIntegerField()

    memory = models.PositiveIntegerField(default=0, help_text='In megabytes')

    used_cores = models.PositiveIntegerField(default=0, help_text='In %')

    used_memory = models.PositiveIntegerField(default=0, help_text='In %')

    info = models.TextField(null=True, blank=True,
                            help_text='Informative message from the worker')


    objects = WorkerManager()


    #_____ Overrides __________


    def __str__(self):
        retval = '%s (%d cores, %d Mb)' % (self.name, self.cores, self.memory)
        if not self.active:
            retval += ' [INACTIVE]'
        return retval


    def natural_key(self):
        return (self.name,)


    def get_admin_change_url(self):
        return reverse('admin:backend_worker_change', args=(self.id,))


    def load(self):
        '''Calculates the number of cores in use or to be used in the future'''
        return sum([j.job.block.queue.cores_per_slot for j in self.splits.all()])


    def current_load(self):
        '''Calculates the number of cores being used currently'''
        return sum([j.job.block.queue.cores_per_slot for j in self.splits.filter(status=JobSplit.PROCESSING)])


    def available_cores(self):
        '''Calculates the number of available cores considering current load'''
        return max(self.cores - self.load(), 0)


    def as_dict(self):
        '''Returns a dictionary-like representation'''
        return dict(cores=self.cores, memory=self.memory)
